# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import os

from os.path import dirname, join, isfile, getsize
from mock import Mock
from ovos_bus_client import Message
from neon_utils.user_utils import get_default_user_config
from neon_minerva.tests.skill_unit_test_base import SkillTestCase


class TestSkill(SkillTestCase):

    def test_00_skill_init(self):
        # Test any parameters expected to be set in init or initialize methods
        from neon_utils.skills.neon_skill import NeonSkill

        self.assertIsInstance(self.skill, NeonSkill)
        self.assertIsInstance(self.skill.support_email, str)

    def test_handle_contact_support(self):
        real_get_support_info = self.skill._get_support_info
        real_parse_attachments = self.skill._parse_attachments
        test_attachments = {'attachment': ''}
        self.skill._get_support_info = Mock(return_value={"test": True})
        self.skill._parse_attachments = Mock(return_value=test_attachments)

        self.skill.ask_yesno = Mock(return_value="no")

        user_config = get_default_user_config()
        user_config["user"]["username"] = "test_user"
        user_config["user"]["email"] = ''
        test_message = Message("test", {"utterance": "This is a test"},
                               {"klat_data": True, "username": "test_user",
                                "user_profiles": [user_config]})
        # Contact Support No Email Declined
        self.skill.handle_contact_support(test_message)
        self.skill.speak_dialog.assert_any_call("no_email", private=True)

        # Contact Support Declined
        test_message.context["user_profiles"][0]["user"]["email"] = \
            "test@neon.ai"
        self.skill.handle_contact_support(test_message)
        self.skill.ask_yesno.assert_called_with("confirm_support",
                                                {"email": "test@neon.ai"})
        self.skill.speak_dialog.assert_called_with("cancelled", private=True)
        # Contact Support Approved No Details
        self.skill.ask_yesno = Mock(return_value="yes")
        self.skill.send_email = Mock()
        self.skill.send_email.return_value = True
        self.skill.handle_contact_support(test_message)
        self.assertEqual(self.skill._get_support_info.call_args[0][0],
                         test_message)
        self.skill.send_email.assert_called_with("Neon AI Diagnostics",
                                                 self.skill._format_email_body(
                                                     {"test": True,
                                                      "user_description": None}
                                                 ),
                                                 test_message, "test@neon.ai",
                                                 attachments=test_attachments)
        self.skill.speak_dialog.assert_called_with("complete",
                                                   {"email": "test@neon.ai"},
                                                   private=True)
        # Contact Support Approved With Details
        self.skill.get_response = Mock(return_value="This is only a test")
        self.skill.handle_contact_support(test_message)
        self.assertEqual(self.skill._get_support_info.call_args[0][0],
                         test_message)
        self.skill.get_response.assert_called_once_with("ask_description",
                                                        num_retries=0)
        self.skill.send_email.assert_called_with(
            "Neon AI Diagnostics", self.skill._format_email_body(
                {"test": True,
                 "user_description": "This is only a test"}),
            test_message, "test@neon.ai", attachments=test_attachments)
        self.skill.speak_dialog.assert_called_with("complete",
                                                   {"email": "test@neon.ai"},
                                                   private=True)

        # Email failed
        self.skill.send_email.return_value = False
        self.skill.handle_contact_support(test_message)
        self.assertEqual(self.skill._get_support_info.call_args[0][0],
                         test_message)
        self.skill.get_response.assert_called_with("ask_description",
                                                   num_retries=0)
        self.skill.send_email.assert_called_with(
            "Neon AI Diagnostics", self.skill._format_email_body(
                {"test": True,
                 "user_description": "This is only a test"}),
            test_message, "test@neon.ai")
        self.skill.speak_dialog.assert_called_with("email_error",
                                                   private=True)

        self.skill._get_support_info = real_get_support_info
        self.skill._parse_attachments = real_parse_attachments

    def test_format_email_body(self):
        test_diagnostics = {"user_profile": "testing",
                            "module_status": {"module": None}}
        # No description
        body = self.skill._format_email_body(test_diagnostics)
        parts = body.split('\n\n')
        self.assertEqual(len(parts), 3)
        self.assertIn(self.skill.support_email, parts[0])
        self.assertEqual(parts[1], "No Description Provided")
        self.assertEqual(parts[2], "- Neon AI")

        # User description provided
        test_diagnostics['user_description'] = "testing diagnostics"
        body = self.skill._format_email_body(test_diagnostics)
        parts = body.split('\n\n')
        self.assertEqual(len(parts), 3)
        self.assertIn(self.skill.support_email, parts[0])
        self.assertEqual(parts[1], test_diagnostics['user_description'])
        self.assertEqual(parts[2], "- Neon AI")

    def test_get_log_files(self):
        from ovos_utils.log import LOG
        test_dir = join(dirname(__file__), "logs")
        real_path = LOG.base_path
        LOG.base_path = test_dir
        logs = self.skill._get_log_files()
        LOG.base_path = real_path
        self.assertEqual(set(logs), {join(test_dir, "audio.log"),
                                     join(test_dir, "skills.log")})
        for log in logs:
            self.assertTrue(isfile(log))
            self.assertEqual(dirname(log), test_dir)

        os.remove(join(test_dir, "neon-utils.log"))

    def test_parse_attachments(self):
        from neon_utils.file_utils import decode_base64_string_to_file
        test_dir = join(dirname(__file__), "logs")
        log_files = [join(test_dir, "audio.log"), join(test_dir, "skills.log")]
        parsed = self.skill._parse_attachments(log_files)
        self.assertEqual(set(parsed.keys()),
                         {'audio_log.txt', 'skills_log.txt'})
        for file, log in parsed.items():
            self.assertIsInstance(log, str)
            test_log_file = join(test_dir, f"test_{file}")
            decode_base64_string_to_file(log, test_log_file)
            with open(test_log_file, 'r') as f:
                test_output = f.read()
            with open(join(test_dir, file.replace('_log.txt',
                                                  '.log')), 'r') as f:
                self.assertEqual(f.read(), test_output)
            os.remove(test_log_file)

        # Test truncated log
        from uuid import uuid4
        test_file = join(test_dir, "truncate.log")
        i = 0
        with open(test_file, 'w+') as f:
            i += 1
            line = str(uuid4()) * 4
            for i in range(100000):
                f.write(f'{i} - {line}\n')
        input_size = int(getsize(test_file))
        self.assertTrue(input_size > 1000000)
        parsed = self.skill._parse_attachments([test_file])
        file = list(parsed.keys())[0]
        log = list(parsed.values())[0]
        self.assertEqual(file, "truncate_log.txt")
        self.assertIsInstance(log, str)
        test_outfile = join(test_dir, "test_truncate.log")
        decode_base64_string_to_file(log, test_outfile)
        with open(test_file, 'r') as f:
            original = f.readlines()
        with open(test_outfile, 'r') as f:
            truncated = f.readlines()
        self.assertEqual(truncated, original[-50000:])
        self.assertLess(getsize(test_outfile), input_size)
        os.remove(test_outfile)
        os.remove(test_file)

    def test_check_service_status(self):
        # TODO
        pass

    def test_get_support_info(self):
        from datetime import datetime
        from neon_utils.net_utils import get_ip_address

        user_config = get_default_user_config()
        user_config["user"]["username"] = "test_user"
        test_context = {"klat_data": True, "username": "test_user",
                        "user_profiles": [user_config]}
        test_message = Message("test", {"utterance": "This is a test"},
                               dict(test_context))
        # No modules respond
        diagnostics = self.skill._get_support_info(test_message)
        diag_time = diagnostics["generated_time_utc"]
        self.assertIsInstance(datetime.fromisoformat(diag_time), datetime)
        pip_info = diagnostics["packages"]
        context = diagnostics["message_context"]
        for key in test_context:
            self.assertEqual(test_context[key], context[key])
        self.assertIsInstance(pip_info, str)
        self.assertEqual(diagnostics, {"user_profile": user_config,
                                       "message_context": context,
                                       "module_status": {"speech": None,
                                                         "audio": None,
                                                         "voice": None,
                                                         "skills": None,
                                                         "gui": None,
                                                         "enclosure": None,
                                                         "admin": None},
                                       "loaded_skills": None,
                                       "host_device": {"ip": get_ip_address()},
                                       "generated_time_utc": diag_time,
                                       "packages": pip_info
                                       })

    def test_get_attachments(self):
        real_status = self.skill._check_service_status
        self.skill._check_service_status = Mock(return_value={'test': True})

        original_context = {"content": "something", "context": True}
        test_message = Message("test", {}, dict(original_context))
        test_profile = {"user": {"username": "test_user"},
                        "data": {"key": "val"}}

        content = self.skill._get_support_info(test_message, test_profile)
        for key in {'user_profile', 'message_context', 'module_status',
                    'loaded_skills', 'packages', 'host_device',
                    'generated_time_utc'}:
            self.assertIn(key, content, key)
        self.assertEqual(content['user_profile'], test_profile)
        self.assertEqual(content['message_context'], original_context)
        self.assertEqual(content['module_status'], {'test': True})
        self.assertIsInstance(content['packages'], str)
        self.assertIsInstance(content['host_device']['ip'], str)
        self.assertIsInstance(content['generated_time_utc'], str)

        self.skill._check_service_status = real_status


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import shutil
import sys
import yaml

from copy import deepcopy
from datetime import datetime
from glob import glob
from os.path import join, basename, getsize, isfile
from subprocess import run
from tempfile import mkdtemp

from ovos_bus_client import Message
from neon_utils.user_utils import get_user_prefs
from neon_utils.skills.neon_skill import NeonSkill
from neon_utils.net_utils import get_ip_address
from neon_utils.file_utils import encode_file_to_base64_string
from neon_utils.parse_utils import validate_email
from ovos_utils import classproperty
from ovos_utils.log import LOG
from ovos_utils.process_utils import RuntimeRequirements
from ovos_workshop.decorators import intent_handler


class SupportSkill(NeonSkill):
    def __init__(self, **kwargs):
        NeonSkill.__init__(self, **kwargs)
        self.extra_diagnostic_files = ['/opt/neon/build_info.json']

    @classproperty
    def runtime_requirements(self):
        return RuntimeRequirements(network_before_load=False,
                                   internet_before_load=False,
                                   gui_before_load=False,
                                   requires_internet=True,
                                   requires_network=True,
                                   requires_gui=False,
                                   no_internet_fallback=False,
                                   no_network_fallback=False,
                                   no_gui_fallback=True)

    @property
    def support_email(self) -> str:
        """
        Email to refer users to for support
        """
        return self.settings.get("support_email") or "support@neon.ai"

    @intent_handler('contact_support.intent')
    def handle_contact_support(self, message: Message):
        """
        Handle a user request to contact support
        :param message: Message associated with request
        """
        user_profile = get_user_prefs(message)
        email_addr = user_profile["user"]["email"]
        if not validate_email(email_addr):
            self.speak_dialog("no_email", private=True)
            email_addr = self.support_email
        if self.ask_yesno("confirm_support",
                          {"email": email_addr}) == "yes":
            if user_profile["response_mode"].get("hesitation"):
                self.speak_dialog("one_moment", private=True)
            diagnostic_info = self._get_support_info(message, user_profile)
            user_description = self.get_response("ask_description",
                                                 num_retries=0)
            diagnostic_info["user_description"] = user_description
            attachment_files = self._parse_attachments(
                self._get_attachments(diagnostic_info))
            if self.send_email(self.translate("email_title"),
                               self._format_email_body(diagnostic_info),
                               message, email_addr,
                               attachments=attachment_files):
                self.speak_dialog("complete",
                                  {"email": email_addr},
                                  private=True)
                return
            LOG.error("Email failed to send, retry without attachments")
            if self.send_email(self.translate("email_title"),
                               self._format_email_body(diagnostic_info),
                               message, email_addr):
                self.speak_dialog("complete",
                                  {"email": email_addr},
                                  private=True)
            else:
                LOG.error(f"Email Failed to send!")
                self.speak_dialog("email_error", private=True)
        else:
            self.speak_dialog("cancelled", private=True)

    @staticmethod
    def _parse_attachments(files: list) -> dict:
        """
        Parse a list of files into a dict of filenames to B64 contents.
        Truncates log files to 50000 lines if they exceed 1MB as an arbitrary
        limit that should safely keep all attachments within email provider
        limits (~10MB-50MB).
        :param files: list of files to include as attachments
        """
        attachments = {}
        for file in files:
            try:
                byte_size = getsize(file)
                if byte_size > 1000000:
                    LOG.info(f"{file} is >1MB, truncating")
                    with open(file, 'r+') as f:
                        lines = f.readlines()
                        f.seek(0)
                        f.writelines(lines[-50000:])
                        f.truncate()

                attachments[basename(file).replace('.log', '_log.txt')] = \
                    encode_file_to_base64_string(file)
            except Exception as e:
                LOG.exception(e)
        return attachments

    def _format_email_body(self, diagnostics: dict) -> str:
        """
        Format the diagnostic data with email dialog and return a string body
        :param diagnostics: diagnostic data to format into the email.
        :returns: email body to send
        """
        return '\n\n'.join((self.translate("email_intro",
                                           {"email": self.support_email}),
                            diagnostics.get('user_description') or
                            "No Description Provided",
                            self.translate("email_signature")))

    def _check_service_status(self, message: Message = None) -> dict:
        """
        Query services on the messagebus and report back their status
        """
        message = message or Message("get_status")

        speech_module = self.bus.wait_for_response(
            message.forward("mycroft.speech.is_ready"))
        speech_status = speech_module.data.get("status") if speech_module \
            else None

        voice_module = self.bus.wait_for_response(
            message.forward("mycroft.voice.is_ready"))
        voice_status = voice_module.data.get("status") if voice_module \
            else None

        audio_module = self.bus.wait_for_response(
            message.forward("mycroft.audio.is_ready"))
        audio_status = audio_module.data.get("status") if audio_module \
            else None

        skills_module = self.bus.wait_for_response(
            message.forward("mycroft.skills.is_ready"))
        skills_status = skills_module.data.get("status") if skills_module \
            else None

        gui_module = self.bus.wait_for_response(
            message.forward("mycroft.gui_service.is_ready"))
        gui_status = gui_module.data.get("status") if gui_module \
            else None

        phal_module = self.bus.wait_for_response(
            message.forward("mycroft.PHAL.is_ready")
        )
        enclosure_status = phal_module.data.get("status") if phal_module \
            else None

        admin_module = self.bus.wait_for_response(
            message.forward("mycroft.PHAL.admin.is_ready")
        )
        admin_status = admin_module.data.get("status") if admin_module \
            else None

        return {"speech": speech_status,
                "voice": voice_status,
                "audio": audio_status,
                "skills": skills_status,
                "gui": gui_status,
                "enclosure": enclosure_status,
                "admin": admin_status}

    @staticmethod
    def _get_log_files():
        log_path = LOG.base_path
        log_files = glob(join(log_path, "*.log"))
        LOG.info(f"Found log files: {log_files}")
        return log_files

    def _get_support_info(self, message: Message,
                          profile: dict = None) -> dict:
        """
        Collect relevant information to include in a support ticket
        :param message: Message associated with support request
        """
        user_profile = profile or get_user_prefs(message)
        message_context = deepcopy(message.context)

        loaded_skills = self.bus.wait_for_response(
            message.forward("skillmanager.list"), "mycroft.skills.list"
        )
        loaded_skills = loaded_skills.data if loaded_skills else None

        core_device_ip = get_ip_address()
        packages = run([sys.executable, "-m", "pip", "list"],
                       capture_output=True).stdout.decode()
        return {
            "user_profile": user_profile,
            "message_context": message_context,
            "module_status": self._check_service_status(message),
            "loaded_skills": loaded_skills,
            "packages": packages,
            "host_device": {"ip": core_device_ip},
            "generated_time_utc": datetime.utcnow().isoformat()
        }

    def _get_attachments(self, info: dict) -> list:
        """
        Parse dict information into attachment files for a support email
        :param info: diagnostic information to parse into attachments
        :returns: list of output attachment files
        """
        log_files = self._get_log_files()
        tempdir = mkdtemp()
        att_files = list()
        # Make a temp copy of log files to optionally truncate before sending
        for log_file in log_files:
            output_file = join(tempdir, basename(log_file))
            shutil.copyfile(log_file, output_file)
            att_files.append(output_file)

        for file in self.extra_diagnostic_files:
            if isfile(file):
                output_file = join(tempdir, basename(file))
                shutil.copyfile(file, output_file)
                att_files.append(output_file)

        packages_file = join(tempdir, "python_packages.txt")
        diagnostics_file = join(tempdir, "diagnostics.txt")
        core_config_file = join(tempdir, "core_config.txt")

        # Add `pip list` output to its own file
        with open(packages_file, 'w+') as f:
            f.write(info.pop('packages', ""))
        att_files.append(packages_file)

        # Add core config output to its own file
        with open(core_config_file, 'w+') as f:
            yaml.dump(dict(self.config_core), f)
        att_files.append(core_config_file)

        # Dump gathered diagnostics to separate file
        with open(diagnostics_file, 'w+') as f:
            yaml.dump(info, f)
        att_files.append(diagnostics_file)

        return att_files

    def stop(self):
        pass
